from tortoise import Model, fields


class User(Model):
    username = fields.CharField(max_length=20, unique=True)
    password = fields.CharField(max_length=200)

    class Meta:
        abstract = True



import importlib
from typing import Type

from fastapi import FastAPI
from tortoise import Model, Tortoise

from .site import Site, Resource, Field


class AdminApp(FastAPI):
    models: str
    admin_secret: str
    user_model: str
    site: Site
    _inited: bool = False
    _field_type_mapping = {
        'IntField': 'number',
        'BooleanField': 'checkbox',
        'DatetimeField': 'datetime',
        'DateField': 'date',
        'IntEnumFieldInstance': 'select',
        'CharEnumFieldInstance': 'select',
        'DecimalField': 'number',
        'FloatField': 'number',
        'TextField': 'textarea',
        'SmallIntField': 'number',
        'ForeignKeyFieldInstance': 'select',
        'JSONField': 'json',
    }
    _model_menu_mapping = {}

    def _get_model_menu_mapping(self):
        for menu in filter(lambda x: x.url, self.site.menus):
            self._model_menu_mapping[menu.url.split('?')[0].split('/')[-1]] = menu

    def init(self, site: Site, user_model: str, admin_secret: str, models: str, ):
        """
        init admin site
        :param site:
        :param user_model: admin user model path,like admin.models.user
        :param admin_secret: admin jwt secret.
        :param models: tortoise models
        :return:
        """
        self.site = site
        self.admin_secret = admin_secret
        self.models = importlib.import_module(models)
        self.user_model = getattr(self.models, user_model)
        self._inited = True
        self._get_model_menu_mapping()

    def _exclude_field(self, resource: str, field: str):
        """
        exclude field by menu include and exclude
        :param resource:
        :param field:
        :return:
        """
        menu = self._model_menu_mapping[resource]
        if menu.include:
            if field not in menu.include:
                return True
        if menu.exclude:
            if field in menu.exclude:
                return True
        return False

    async def _build_resource_from_model_describe(self, resource: str, model: Type[Model], model_describe: dict,
                                                  exclude_readonly: bool):
        data_fields = model_describe.get('data_fields')
        pk_field = model_describe.get('pk_field')
        fk_fields = model_describe.get('fk_fields')
        menu = self._model_menu_mapping[resource]
        search_fields_ret = {}
        search_fields = menu.search_fields
        if exclude_readonly:
            fields = {}
        else:
            fields = {
                pk_field.get('name'): Field(
                    label=pk_field.get('name').title(),
                    required=True,
                    type=self._field_type_mapping.get(pk_field.get('field_type')) or 'text',
                )
            }
        for field in data_fields:
            read_only = field.get('constraints').get('readOnly')
            field_type = field.get('field_type')
            name = field.get('name')

            if (read_only and exclude_readonly) or self._exclude_field(resource, name):
                continue

            type_ = self._field_type_mapping.get(field_type) or 'text'
            options = []
            if type_ == 'select':
                for k, v in model._meta.fields_map[name].enum_type.choices().items():
                    options.append({'text': v, 'value': k})

            label = field.get('description') or field.get('name').title()
            fields[name] = Field(
                label=label,
                required=not field.get('nullable'),
                type=type_,
                options=options
            )
            if name in search_fields:
                search_fields_ret[name] = Field(
                    label=label,
                )

        for fk_field in fk_fields:
            name = fk_field.get('name')
            if name not in menu.raw_id_fields:
                fk_model_class = model._meta.fields_map[name].model_class
                objs = await fk_model_class.all()
                raw_field = fk_field.get('raw_field')
                label = fk_field.get('description') or name.title()
                options = list(map(lambda x: {'text': str(x), 'value': x.pk}, objs))
                fields[raw_field] = Field(
                    label=label,
                    required=True,
                    type='select',
                    options=options
                )
                if name in search_fields:
                    search_fields_ret[raw_field] = Field(
                        label=label,
                        type='select',
                        options=options,
                    )
        return fields, search_fields_ret

    async def get_resource(self, resource: str, exclude_readonly=False, ):
        assert self._inited, 'must call init() first!'
        model = getattr(self.models, resource)  # type:Type[Model]
        model_describe = Tortoise.describe_model(model)
        fields, search_fields = await self._build_resource_from_model_describe(resource, model, model_describe,
                                                                               exclude_readonly)
        return Resource(
            title=model_describe.get('description') or resource.title(),
            fields=fields,
            searchFields=search_fields
        )


app = AdminApp(
    openapi_prefix='/admin',
)



from typing import Optional, Tuple, Dict, Union, List, Set

from pydantic import BaseModel, HttpUrl


class Menu(BaseModel):
    name: str
    title: Optional[bool]
    url: Optional[str]
    icon: Optional[str]
    include: Optional[Tuple[str]]
    exclude: Optional[Tuple[str]]
    external: Optional[bool] = False
    raw_id_fields: Optional[Set[str]] = set()
    search_fields: Optional[Set[str]] = set()


class Site(BaseModel):
    name: str
    logo: HttpUrl
    locale: str
    locale_switcher: bool
    menus: List[Menu]

    class Config:
        fields = {
            'menus': 'menu'
        }


class Field(BaseModel):
    label: str
    cols: Optional[int]
    input_cols: Optional[int]
    group: Optional[str]
    type: Optional[Union[str, Dict]]
    required: bool = True
    options: Optional[List[Dict]]


class Resource(BaseModel):
    title: str
    resource_fields: Dict[str, Field]
    searchFields: Optional[Dict[str, Field]]

    class Config:
        fields = {
            'resource_fields': 'fields',
        }



from fastapi import Depends, APIRouter
from fastapi.responses import ORJSONResponse
from starlette.status import HTTP_409_CONFLICT
from tortoise.contrib.pydantic import pydantic_model_creator
from tortoise.exceptions import IntegrityError

from ..depends import QueryItem, get_query, parse_body, get_model
from ..factory import app
from ..responses import GetManyOut
from ..shortcuts import get_object_or_404

router = APIRouter()


@router.get(
    '/{resource}'
)
async def get_resource(
        resource: str,
        query: QueryItem = Depends(get_query),
        model=Depends(get_model)
):
    qs = model.all()
    if query.where:
        qs = qs.filter(**query.where)
    result = await qs.limit(query.size).offset((query.page - 1) * query.size)
    resource = await app.get_resource(resource)
    creator = pydantic_model_creator(model, include=resource.resource_fields.keys())
    return GetManyOut(
        total=await qs.count(),
        data=list(map(lambda x: creator.from_orm(x).dict(), result))
    )


@router.get(
    '/{resource}/form'
)
async def form(
        resource: str,
):
    resource = await app.get_resource(resource, exclude_readonly=True)
    return resource.dict(by_alias=True, exclude_unset=True)


@router.get(
    '/{resource}/grid'
)
async def grid(
        resource: str,
):
    resource = await app.get_resource(resource)
    return resource.dict(by_alias=True, exclude_unset=True)


@router.get(
    '/{resource}/view'
)
async def view(
        resource: str,
):
    resource = await app.get_resource(resource)
    return resource.dict(by_alias=True, exclude_unset=True)


@router.delete(
    '/{resource}/{id}'
)
async def delete_one(
        id: int,
        model=Depends(get_model)
):
    await model.filter(pk=id).delete()
    return {'success': True}


@router.put(
    '/{resource}/{id}'
)
async def update_one(
        id: int,
        parsed=Depends(parse_body),
        model=Depends(get_model)
):
    body, resource_fields = parsed
    try:
        print(body)
        await model.filter(pk=id).update(**body)
    except IntegrityError as e:
        return ORJSONResponse(status_code=HTTP_409_CONFLICT, content=dict(
            message=f'Update Error,{e}'
        ))
    user_ = await get_object_or_404(model, pk=id)
    creator = pydantic_model_creator(model, include=resource_fields)
    return creator.from_orm(user_).dict()


@router.post(
    '/{resource}'
)
async def create_one(
        parsed=Depends(parse_body),
        model=Depends(get_model)
):
    body, resource_fields = parsed
    creator = pydantic_model_creator(model, include=resource_fields)
    try:
        obj = await model.create(**body)
    except IntegrityError as e:
        return ORJSONResponse(status_code=HTTP_409_CONFLICT, content=dict(
            message=f'Create Error,{e}'
        ))
    return creator.from_orm(obj).dict()


@router.get(
    '/{resource}/{id}'
)
async def get_one(
        id: int,
        resource: str,
        model=Depends(get_model)
):
    obj = await get_object_or_404(model, pk=id)
    resource = await app.get_resource(resource)
    include = resource.resource_fields.keys()
    creator = pydantic_model_creator(model, include=include)
    return creator.from_orm(obj).dict()



import jwt
from fastapi import APIRouter
from fastapi.responses import ORJSONResponse
from passlib.context import CryptContext
from starlette.status import HTTP_403_FORBIDDEN

from ..factory import app
from ..schemas import LoginIn
from ..shortcuts import get_object_or_404

pwd_context = CryptContext(schemes=['bcrypt'], deprecated='auto')

router = APIRouter()


@router.post(
    '/login',
)
async def login(
        login_in: LoginIn
):
    user_model = app.user_model
    user = await get_object_or_404(user_model, username=login_in.username)
    if not pwd_context.verify(login_in.password, user.password):
        return ORJSONResponse(status_code=HTTP_403_FORBIDDEN, content={
            'message': 'Incorrect Password'
        })
    return {
        'user': {
            'username': user.username
        },
        'token': jwt.encode({'user_id': user.pk}, app.admin_secret, algorithm='HS256')
    }



from fastapi import Depends

from . import login, other, index
from ..depends import jwt_required
from ..factory import app

app.include_router(login.router)
app.include_router(other.router, dependencies=[Depends(jwt_required)])
app.include_router(index.router, dependencies=[Depends(jwt_required)])



import jwt
import orjson
from fastapi import Query, Path, Depends, HTTPException
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from pydantic import BaseModel
from starlette.requests import Request
from starlette.status import HTTP_401_UNAUTHORIZED
from .factory import app

auth_schema = HTTPBearer()


async def jwt_required(request: Request, token: HTTPAuthorizationCredentials = Depends(auth_schema)):
    credentials_exception = HTTPException(HTTP_401_UNAUTHORIZED)
    try:
        payload = jwt.decode(token.credentials, app.admin_secret)
        user_id = payload.get('user_id')
        if user_id is None:
            raise credentials_exception
    except jwt.PyJWTError:
        raise credentials_exception
    request.scope['user_id'] = user_id
    return user_id


class QueryItem(BaseModel):
    page: int = 1
    sort: dict
    where: dict
    with_: dict
    size: int = 10

    class Config:
        fields = {
            'with_': 'with'
        }


def get_query(query=Query(...)):
    query = orjson.loads(query)
    return QueryItem.parse_obj(query)


def get_model(resource: str = Path(...)):
    model = getattr(app.models, resource)
    return model


async def parse_body(request: Request, resource: str = Path(...)):
    body = await request.json()
    resource = await app.get_resource(resource, exclude_readonly=True)
    resource_fields = resource.resource_fields.keys()
    ret = {}
    for key in resource_fields:
        ret[key] = body.get(key)
    return ret, resource_fields



from typing import Sequence, Dict
from pydantic import BaseModel


class GetManyOut(BaseModel):
    total: int
    per_page: int = 10
    page: int = 1
    data: Sequence[Dict]

    class Config:
        fields = {
            'per_page': 'perPage'
        }



import os

from setuptools import find_packages, setup
from fastapi_admin import __version__

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    long_description = f.read()


def requirements():
    return open('requirements.txt', 'rt').read().splitlines()


setup(
    name='fastapi-admin',
    version=__version__,
    description='Fast Admin Dashboard based on fastapi and tortoise-orm and rest-admin.',
    author='long2ice',
    long_description_content_type='text/x-rst',
    long_description=long_description,
    author_email='long2ice@gmail.com',
    url='https://github.com/long2ice/fastapi-admin',
    license='MIT License',
    packages=find_packages(include=['fastapi_admin*']),
    include_package_data=True,
    platforms='any',
    keywords=(
        'fastapi admin dashboard'
    ),
    dependency_links=['https://github.com/long2ice/tortoise-orm.git@long2ice#egg=tortoise-orm'],
    install_requires=requirements(),
)



from typing import Generic

from fastapi import HTTPException
from starlette.status import HTTP_404_NOT_FOUND
from tortoise.models import MODEL


async def get_object_or_404(model: Generic[MODEL], **kwargs):
    """
    get_object_or_404
    :param model:
    :param kwargs:
    :return:
    """
    obj = await model.filter(**kwargs).first()  # type:model
    if not obj:
        raise HTTPException(HTTP_404_NOT_FOUND, 'Not Found')
    return obj



from fastapi import APIRouter

from ..factory import app

router = APIRouter()


@router.get(
    '/home',
)
async def home():
    return {
        'title': "Welcome to REST ADMIN"
    }


@router.get(
    '/site',
)
async def site():
    return app.site.dict(by_alias=True, exclude_unset=True)
